import logging
from collections import defaultdict
from typing import Dict, Generator, Iterator, List, Optional, Tuple, Union

from django.contrib.contenttypes.models import ContentType
from rest_framework import filters
from rest_framework.pagination import CursorPagination, LimitOffsetPagination
from rest_framework.request import Request
from rest_framework.reverse import reverse

from bridger.enums import Button, WidgetType
from bridger.filters import DjangoFilterBackend
from bridger.serializers import ListSerializer, RepresentationSerializer

from .metadata import BridgerMetaData
from .utils import ilen

logger = logging.getLogger(__name__)

# FIELD         --> Default Value
# get_field     --> Default Implementation
# _get_field    --> Transformation


class MetadataMixin(
    IdentifierMetadataMixin, WidgetTypeMetadataMixin, EndpointMetadataMixin
):
    metadata_class = BridgerMetaData

    def get_instance_display(self, request: Request) -> List:
        if hasattr(self, "INSTANCE_DISPLAY"):
            return self.INSTANCE_DISPLAY.to_dict()
        return []

    def get_list_display(self, request: Request) -> Dict:
        if hasattr(self, "LIST_DISPLAY"):
            return self.LIST_DISPLAY.to_dict()
        return {}

    def get_fields(self, request: Request) -> Dict:
        fields = dict()
        rs = RepresentationSerializer
        ls = ListSerializer
        field_items = self.get_serializer().fields.items()

        for name, field in filter(
            lambda f: not isinstance(f[1], (rs, ls)), field_items
        ):
            fields[name] = field.get_representation(request, name)

        for name, field in filter(lambda f: isinstance(f[1], (rs, ls)), field_items):
            representation = field.get_representation(request, name)
            fields[representation["related_key"]].update(representation)
            del fields[representation["related_key"]]["related_key"]

        return fields

    def get_preview_display(self, request: Request):
        return getattr(self, "PREVIEW_DISPLAY", "")

    def get_preview_buttons(self, request: Request):
        return [button.to_dict() for button in getattr(self, "PREVIEW_BUTTONS", [])]



from collections import defaultdict

from rest_framework.metadata import BaseMetadata
from rest_framework.reverse import reverse

from .enums import WidgetType
from .serializers import percent_decorator

# metadata["custom_buttons"] = view.get_custom_buttons(request=request)
# metadata["custom_instance_buttons"] = view.get_custom_instance_buttons(
#     request=request
# )

# metadata["preview_buttons"] = view.get_preview_buttons(request=request)
# metadata["preview_display"] = view.get_preview_display(request=request)
# metadata["titles"] = view.get_titles(request=request)

# if metadata["type"] in [WidgetType.INSTANCE.value, WidgetType.LIST.value]:
#     serializer_class = view.get_serializer_class()
#     if "pk" in view.kwargs:
#         metadata["pk"] = view.kwargs["pk"]

#     metadata["list_display"] = view.get_list_display(request)
#     metadata["instance_display"] = view.get_instance_display(request)

#     metadata["fields"] = view.get_fields(request)
#     for key, value in serializer_class.get_decorators():
#         metadata["fields"][key]["decorators"].append(value)
#     for key in serializer_class.get_percent_fields():
#         metadata["fields"][key]["type"] = "percent"
#         metadata["fields"][key]["decorators"].append(percent_decorator)

# metadata["search_fields"] = view.get_search_fields(request)
# metadata["ordering_fields"] = view.get_ordering_fields(request)
# metadata["filter_fields"] = {k: v for k, v in view.get_filter_fields(request)}

# # TODO: Custom Buttons

# return metadata



from typing import Dict, Iterator, List, Tuple, Union

from rest_framework.request import Request
from rest_framework.reverse import reverse

from bridger.metadata.mixins import BridgerMetadataMixin

Endpoint_T = Union[str, Tuple[str, List, Dict]]


class EndpointMetadata(BridgerMetadataMixin):
    key = "endpoints"
    method_name = "_get_endpoints"


class EndpointMetadataMixin:
    def get_generic_endpoint(
        self, request: Request, endpoint: Endpoint_T = None
    ) -> Endpoint_T:
        return endpoint

    def _get_generic_endpoint(self, request: Request, endpoint_type: str) -> Endpoint_T:
        field_name = endpoint_type.upper()
        method_name = f"get_{endpoint_type}"

        endpoint = getattr(self, method_name, self.get_generic_endpoint)(
            request, getattr(self, field_name, None)
        )

        if endpoint is None:
            return endpoint

        if isinstance(endpoint, str):
            viewname, args, kwargs = endpoint, [], {}
        else:
            viewname, args, kwargs = endpoint

        return reverse(viewname=viewname, request=request, args=args, kwargs=kwargs)

    def _get_endpoint(self, request: Request) -> Endpoint_T:
        return self._get_generic_endpoint(request, "endpoint")

    def _get_list_endpoint(self, request: Request) -> Endpoint_T:
        return self._get_generic_endpoint(request, "list_endpoint")

    def _get_instance_endpoint(self, request: Request) -> Endpoint_T:
        return self._get_generic_endpoint(request, "instance_endpoint")

    def _get_create_endpoint(self, request: Request) -> Endpoint_T:
        return self._get_generic_endpoint(request, "create_endpoint")

    def _get_delete_endpoint(self, request: Request) -> Endpoint_T:
        return self._get_generic_endpoint(request, "delete_endpoint")

    def _get_endpoints(self, request: Request) -> Iterator[Dict[str, str]]:
        endpoints = dict()

        endpoint = self._get_endpoint(request=request)
        list_endpoint = self._get_list_endpoint(request=request)
        instance_endpoint = self._get_instance_endpoint(request=request)
        create_endpoint = self._get_create_endpoint(request=request)
        delete_endpoint = self._get_delete_endpoint(request=request)

        if list_endpoint or endpoint:
            endpoints["list"] = list_endpoint or endpoint

        if instance_endpoint or endpoint:
            endpoints["instance"] = instance_endpoint or endpoint

        if create_endpoint or endpoint:
            endpoints["create"] = create_endpoint or endpoint

        if delete_endpoint or endpoint:
            endpoints["delete"] = delete_endpoint or endpoint

        return endpoints



from dataclasses import dataclass, field
from typing import Any, List, Optional

from bridger.enums import RequestType

from .bases import ButtonConfig, ButtonTypeMixin, ButtonUrlMixin
from .enums import ButtonType, HyperlinkTarget


@dataclass
class DropDownButton(ButtonTypeMixin, ButtonConfig):
    button_type = ButtonType.DROPDOWN
    buttons: List = field(default_factory=list)

    def __iter__(self):
        yield from super().__iter__()
        yield "buttons", [dict(button) for button in self.buttons]


@dataclass
class WidgetButton(ButtonTypeMixin, ButtonUrlMixin, ButtonConfig):
    button_type = ButtonType.WIDGET
    new_mode: bool = False

    def __iter__(self):
        yield from super().__iter__()
        yield "new_mode", self.new_mode


@dataclass
class HyperlinkButton(ButtonTypeMixin, ButtonUrlMixin, ButtonConfig):
    button_type = ButtonType.HYPERLINK
    target: HyperlinkTarget = HyperlinkTarget.BLANK

    def __iter__(self):
        yield from super().__iter__()
        yield "target", self.target.value


@dataclass
class ActionButton(ButtonTypeMixin, ButtonUrlMixin, ButtonConfig):
    button_type = ButtonType.ACTION
    method: RequestType = RequestType.POST
    action_label: str = ""

    description_fields: str = "<p>Are you sure you want to proceed?</p>"
    confirm_config: ButtonConfig = ButtonConfig(label="Confirm", title="Confirm")
    cancel_config: ButtonConfig = ButtonConfig(label="Cancel", title="Cancel")

    identifiers: List[str] = field(default_factory=list)

    # TODO:
    # Instance Display
    # Fields
    # This still has to be worked through:
    instance_display: Any = None
    fields: Any = None

    def __iter__(self):
        yield from super().__iter__()
        yield "action_label", self.action_label
        yield "method", self.method.value

        yield "description_fields", self.description_fields
        yield "confirm_config", dict(self.confirm_config)
        yield "cancel_config", dict(self.cancel_config)

        yield "identifiers", self.identifiers
